@@ -35,6 +35,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <sys/stat.h>
+#include <sys/sysmacros.h>
 #include <systemd/sd-login.h>
 #include <unistd.h>
 
@@ -837,12 +838,12 @@ launcher_logind_get_vt(struct weston_launcher *launcher)
 	return wl->vtnr;
 }
 
-struct launcher_interface launcher_logind_iface = {
-	launcher_logind_connect,
-	launcher_logind_destroy,
-	launcher_logind_open,
-	launcher_logind_close,
-	launcher_logind_activate_vt,
-	launcher_logind_restore,
-	launcher_logind_get_vt,
+const struct launcher_interface launcher_logind_iface = {
+	.connect = launcher_logind_connect,
+	.destroy = launcher_logind_destroy,
+	.open = launcher_logind_open,
+	.close = launcher_logind_close,
+	.activate_vt = launcher_logind_activate_vt,
+	.restore = launcher_logind_restore,
+	.get_vt = launcher_logind_get_vt,
 };
